@@ -26,7 +26,7 @@ with ods.open('imas', os.environ['USER'], 'DIII-D', 1000, 0):
     # then it is in memory
     print(ods['equilibrium.time_slice.0.global_quantities.ip'])
 
-# save it as a pickle (which will keep memory of the dynamic nature of the ODS)
+# save it as a pickle
 ods.save('test.pkl')
 
 # load the data back
@@ -40,7 +40,7 @@ print(ods.flat().keys())
 with ods.open('imas', os.environ['USER'], 'DIII-D', 1000, 0):
     print(ods['equilibrium.time'])
 
-# dynamic load can also tell us what IMAS elements have data
+# tell us what IMAS elements have data
 with ods.open('imas', os.environ['USER'], 'DIII-D', 1000, 0):
-    print(ods.keys(dynamic=True))
-    print(ods['equilibrium.time_slice.0.profiles_1d'].keys(dynamic=True))
+    print(ods.keys())
+    print(ods['equilibrium.time_slice.0.profiles_1d'].keys())
