@@ -294,8 +294,33 @@ def pf_active_hardware(ods, pulse):
     mhdin = get_support_file(OMFITmhdin, support_filenames('d3d', 'mhdin', pulse))
     mhdin.to_omas(ods, update='pf_active')
 
-    for k in range(len(ods['pf_active.coil'])):
-        fcid = 'F{}{}'.format((k % 9) + 1, 'AB'[int(mhdin['FC'][k, 1] < 0)])
+    coil_names = [
+        'F1A',
+        'F2A',
+        'F3A',
+        'F4A',
+        'F5A',
+        'F6A',
+        'F7A',
+        'F8A',
+        'F9A',
+        'F2B',
+        'F2B',
+        'F3B',
+        'F4B',
+        'F5B',
+        'F6B',
+        'F7B',
+        'F8B',
+        'F9B',
+        'ECOILA',
+        'ECOILB',
+        'E567UP',
+        'E567DN',
+        'E89DN',
+        'E89UP',
+    ]
+    for k, fcid in enumerate(coil_names):
         ods['pf_active.coil'][k]['name'] = fcid
         ods['pf_active.coil'][k]['identifier'] = fcid
         ods['pf_active.coil'][k]['element.0.name'] = fcid
@@ -325,6 +350,71 @@ def pf_active_coil_current_data(ods, pulse):
             tdi_expression='ptdata2("{signal}",{pulse})',
             time_norm=0.001,
             data_norm=1.0,
+            homogeneous_time=False,
+        )
+
+        # fetch uncertainties
+        TDIs = {}
+        for k in ods1['pf_active.coil']:
+            identifier = ods1[f'pf_active.coil.{k}.identifier'].upper()
+            TDIs[identifier] = f'pthead2("{identifier}",{pulse}), __rarray'
+
+        data = mdsvalue('d3d', None, pulse, TDIs).raw()
+        for k in ods1['pf_active.coil']:
+            identifier = ods1[f'pf_active.coil.{k}.identifier'].upper()
+            nt = len(ods[f'pf_active.coil.{k}.current.data'])
+            ods[f'pf_active.coil.{k}.current.data_error_upper'] = abs(data[identifier][3] * data[identifier][4]) * np.ones(nt) * 10.0
+
+
+# ================================
+@machine_mapping_function(__regression_arguments__)
+def coils_non_axisymmetric_hardware(ods, pulse):
+    r"""
+    Loads DIII-D tokamak non_axisymmetric field coil hardware geometry
+
+    :param ods: ODS instance
+    """
+
+    from omfit_classes.omfit_omas_d3d import OMFITd3dcompfile
+
+    coil_names = []
+    for compfile in ['ccomp', 'icomp']:
+        comp = get_support_file(OMFITd3dcompfile, support_filenames('d3d', compfile, pulse))
+        compshot = -1
+        for shot in comp:
+            if pulse > compshot:
+                compshot = shot
+
+        coil_names += list(comp[compshot].keys())
+
+    for k, fcid in enumerate(coil_names):
+        ods['coils_non_axisymmetric.coil'][k]['name'] = fcid
+        ods['coils_non_axisymmetric.coil'][k]['identifier'] = fcid
+
+
+@machine_mapping_function(__regression_arguments__, pulse=133221)
+def coils_non_axisymmetric_current_data(ods, pulse):
+    # get pf_active hardware description --without-- placing the data in this ods
+    # use `unwrap` to avoid calling `@machine_mapping_function` of `pf_active_hardware`
+    ods1 = ODS()
+    unwrap(coils_non_axisymmetric_hardware)(ods1, pulse)
+
+    # fetch the actual pf_active currents data
+    with omas_environment(ods):
+        fetch_assign(
+            ods,
+            ods1,
+            pulse,
+            channels='coils_non_axisymmetric.coil',
+            identifier='coils_non_axisymmetric.coil.{channel}.identifier',
+            time='coils_non_axisymmetric.coil.{channel}.current.time',
+            data='coils_non_axisymmetric.coil.{channel}.current.data',
+            validity=None,
+            mds_server='d3d',
+            mds_tree='D3D',
+            tdi_expression='ptdata2("{signal}",{pulse})',
+            time_norm=0.001,
+            data_norm=1.0,
         )
 
 
@@ -362,40 +452,43 @@ def ec_launcher_active_hardware(ods, pulse):
             if field in ['XMFRAC', 'FPWRC', 'AZIANG', 'POLANG']:
                 query["TIME_" + field + f'_{system_no}'] = "dim_of(" + query[field + f'_{system_no}'] + "+01)"
     # Final, third query now that we have resolved all the TDIs related to gyrotron names
-    gyortrons = mdsvalue('d3d', treename='RF', pulse=pulse, TDI=query).raw()
+    gyrotrons = mdsvalue('d3d', treename='RF', pulse=pulse, TDI=query).raw()
     for system_no in range(1, num_systems + 1):
         system_index = system_no - 1
-        if gyortrons[f'STAT_{system_no}'] == 0:
+        if gyrotrons[f'STAT_{system_no}'] == 0:
             continue
-        launchers = ods['ec_launchers.launcher']
-        launchers[system_index]['identifier'] = launchers[system_index]['name'] = gyrotron_names[system_index]
-        launchers[system_index]['launching_position.r'] = np.atleast_1d(systems[f'LAUNCH_R_{system_no}'])
-        launchers[system_index]['launching_position.z'] = np.atleast_1d(systems[f'LAUNCH_Z_{system_no}'])
-        launchers[system_index]['launching_position.time'] = np.zeros(launchers[system_index]['launching_position.r'].shape)
+        beams = ods['ec_launchers.beam']
+        time = np.atleast_1d(gyrotrons[f'TIME_AZIANG_{system_no}']) / 1.0e3
+        if len(time) == 1:
+            beams[system_index]['time'] = np.atleast_1d(0)
+        else:
+            beams[system_index]['time'] = np.atleast_1d(gyrotrons[f'TIME_AZIANG_{system_no}']) / 1.0e3
+        ntime = len(beams[system_index]['time'])
+        beams[system_index]['steering_angle_tor'] = np.atleast_1d(np.deg2rad((gyrotrons[f'AZIANG_{system_no}'] - 180.0)))
+        beams[system_index]['steering_angle_pol'] = np.atleast_1d(np.deg2rad((gyrotrons[f'POLANG_{system_no}'] - 90.0)))
+
+        beams[system_index]['identifier'] = beams[system_index]['name'] = gyrotron_names[system_index]
+
+        beams[system_index]['launching_position.r'] = np.atleast_1d(systems[f'LAUNCH_R_{system_no}'])[0] * np.ones(ntime)
+        beams[system_index]['launching_position.z'] = np.atleast_1d(systems[f'LAUNCH_Z_{system_no}'])[0] * np.ones(ntime)
+
         phi = np.deg2rad(float(systems[f'PORT_{system_no}'].split(' ')[0]))
         phi = -phi - np.pi / 2.0
-        launchers[system_index]['launching_position.phi'] = np.atleast_1d(phi)
-        launchers[system_index]['frequency.data'] = np.atleast_1d(systems[f'FREQUENCY_{system_no}'])
-        launchers[system_index]['frequency.time'] = np.atleast_1d(0)
-        launchers[system_index]['power_launched.time'] = np.atleast_1d(gyortrons[f'TIME_FPWRC_{system_no}']) / 1.0e3
-        launchers[system_index]['power_launched.data'] = np.atleast_1d(gyortrons[f'FPWRC_{system_no}'])
-        xfrac = gyortrons[f'XMFRAC_{system_no}']
+        beams[system_index]['launching_position.phi'] = phi * np.ones(ntime)
+
+        beams[system_index]['frequency.data'] = np.atleast_1d(systems[f'FREQUENCY_{system_no}'])
+        beams[system_index]['frequency.time'] = np.atleast_1d(0)
+
+        beams[system_index]['power_launched.time'] = np.atleast_1d(gyrotrons[f'TIME_FPWRC_{system_no}']) / 1.0e3
+        beams[system_index]['power_launched.data'] = np.atleast_1d(gyrotrons[f'FPWRC_{system_no}'])
+
+        xfrac = gyrotrons[f'XMFRAC_{system_no}']
+
         if iterable(xfrac):
-            launchers[system_index]['mode.data'] = np.atleast_1d(np.array(np.round(1.0 - 2.0 * xfrac), dtype=np.int))
-            launchers[system_index]['mode.time'] = np.atleast_1d(gyortrons[f'TIME_XMFRAC_{system_no}']) / 1.0e3
+            beams[system_index]['mode'] = int(np.round(1.0 - 2.0 * xfrac)[0])
         else:
-            launchers[system_index]['mode.data'] = np.atleast_1d([np.round(1.0 - 2.0 * xfrac)], dtype=np.int)
-            launchers[system_index]['mode.time'] = np.atleast_1d(0)
-        launchers[system_index]['steering_angle_tor.data'] = np.atleast_1d(np.deg2rad((gyortrons[f'AZIANG_{system_no}'] - 180.0)))
-        if len(launchers[system_index]['steering_angle_tor.data']) == 1:
-            launchers[system_index]['steering_angle_tor.time'] = np.atleast_1d(0)
-        else:
-            launchers[system_index]['steering_angle_tor.time'] = np.atleast_1d(gyortrons[f'TIME_AZIANG_{system_no}']) / 1.0e3
-        launchers[system_index]['steering_angle_pol.data'] = np.atleast_1d(np.deg2rad((gyortrons[f'POLANG_{system_no}'] - 90.0)))
-        if len(launchers[system_index]['steering_angle_pol.data']) == 1:
-            launchers[system_index]['steering_angle_pol.time'] = np.atleast_1d(0)
-        else:
-            launchers[system_index]['steering_angle_pol.time'] = np.atleast_1d(gyortrons[f'TIME_POLANG_{system_no}']) / 1.0e3
+            beams[system_index]['mode'] = int(np.round(1.0 - 2.0 * xfrac))
+
         # The spot size and radius are computed using the evolution formula for Gaussian beams
         # see: https://en.wikipedia.org/wiki/Gaussian_beam#Beam_waist
         # The beam is divergent because the beam waist is focused on to the final launching mirror.
@@ -403,14 +496,10 @@ def ec_launcher_active_hardware(ods, pulse):
         # the beam is focused onto the mirror meaning that it is paraxial at the launching point.
         # Hence, the inverse curvature radius is zero
         # Notably the ECRH beams are astigmatic in reality so this is just an approximation
-        launchers[system_index]['beam.phase.angle.time'] = np.array([0.0])
-        launchers[system_index]['beam.phase.angle.data'] = np.deg2rad([0.0])
-        launchers[system_index]['beam.phase.curvature.time'] = np.array([0.0])
-        launchers[system_index]['beam.phase.curvature.data'] = np.array([np.atleast_1d(0.0), np.atleast_1d(0.0)])
-        launchers[system_index]['beam.spot.angle.time'] = np.array([0.0])
-        launchers[system_index]['beam.spot.angle.data'] = np.deg2rad([0.0])
-        launchers[system_index]['beam.spot.size.time'] = np.array([0.0])
-        launchers[system_index]['beam.spot.size.data'] = np.array([np.atleast_1d(0.0172), np.atleast_1d(0.0172)])
+        beams[system_index]['phase.angle'] = np.zeros(ntime)
+        beams[system_index]['phase.curvature'] = np.zeros([2, ntime])
+        beams[system_index]['spot.angle'] = np.zeros(ntime)
+        beams[system_index]['spot.size'] = 0.0172 * np.ones([2, ntime])
 
 
 # ================================
@@ -419,7 +508,8 @@ def interferometer_hardware(ods, pulse):
     """
     Loads DIII-D CO2 interferometer chord locations
 
-    The chord endpoints ARE NOT RIGHT. Only the R for vertical lines or Z for horizontal lines is right.
+    The chord endpoints are approximative. They do not take into account 
+    the vessel wall contour of the shot. The values have been taken from OMFITprofiles.
 
     Data sources:
     DIII-D webpage: https://diii-d.gat.com/diii-d/Mci accessed 2018 June 07 by D. Eldon
@@ -434,15 +524,16 @@ def interferometer_hardware(ods, pulse):
     ods['interferometer.channel.0.identifier'] = ods['interferometer.channel.0.name'] = 'r0'
     r0 = ods['interferometer.channel.0.line_of_sight']
     r0['first_point.phi'] = r0['second_point.phi'] = 225 * (-np.pi / 180.0)
-    r0['first_point.r'], r0['second_point.r'] = 3.0, 0.8  # These are not the real endpoints
+    r0['first_point.r'], r0['second_point.r'] = 2.36, 1.01  # End points from OMFITprofiles
     r0['first_point.z'] = r0['second_point.z'] = 0.0
-
+    Z_top = 1.24
+    Z_bottom = -1.375
     for i, r in enumerate([1.48, 1.94, 2.10]):
         ods['interferometer.channel'][i + 1]['identifier'] = ods['interferometer.channel'][i + 1]['name'] = 'v{}'.format(i + 1)
         los = ods['interferometer.channel'][i + 1]['line_of_sight']
         los['first_point.phi'] = los['second_point.phi'] = 240 * (-np.pi / 180.0)
         los['first_point.r'] = los['second_point.r'] = r
-        los['first_point.z'], los['second_point.z'] = -1.8, 1.8  # These are not the real points
+        los['first_point.z'], los['second_point.z'] = Z_top, Z_bottom  # End points from OMFITprofiles
 
     for i in range(len(ods['interferometer.channel'])):
         ch = ods['interferometer.channel'][i]
@@ -889,7 +980,7 @@ def langmuir_probes_data(ods, pulse, _get_measurements=True):
                         area=1e-4,  # cm^2 --> m^2
                         pot=1,  # V --> V
                         angle=np.pi / 180,  # degrees --> radians
-                        heatflux=1e3 * 1e-4,  # kW cm^-2 --> W m^-2
+                        heatflux=1e3 * 1e4,  # kW cm^-2 --> W m^-2
                     )
                     for tdi_part, imas_part in nodes.items():
                         mds_dat = mdsvalue('d3d', pulse=pulse, treename='langmuir', TDI=rf'\langmuir::top.probe_{i:03d}.{tdi_part}')
@@ -964,6 +1055,24 @@ def charge_exchange_data(ods, pulse, analysis_type='CERQUICK', _measurements=Tru
 
 
 # ================================
+@machine_mapping_function(__regression_arguments__)
+def magnetics_weights(ods, pulse, time_index):
+    r"""
+    Load DIII-D tokamak magnetics equilibrium weights
+
+    :param ods: ODS instance
+    """
+    from omfit_classes.omfit_omas_d3d import OMFITd3dfitweight
+
+    fitweight = get_support_file(OMFITd3dfitweight, support_filenames('d3d', 'fitweight', pulse))
+    weight_ishot = -1
+    for ishot in fitweight:
+        if pulse > ishot and ishot > weight_ishot:
+            weight_ishot = ishot
+    ods['equilibrium.time_slice.{time_index}.constraints.bpol_probe.:.weight'] = fitweight[weight_ishot]['fwtmp2']
+    ods['equilibrium.time_slice.{time_index}.constraints.flux_loop.:.weight'] = fitweight[weight_ishot]['fwtsi']
+
+
 @machine_mapping_function(__regression_arguments__)
 def magnetics_hardware(ods, pulse):
     r"""
@@ -978,7 +1087,10 @@ def magnetics_hardware(ods, pulse):
 
 
 @machine_mapping_function(__regression_arguments__, pulse=133221)
-def magnetics_floops_data(ods, pulse):
+def magnetics_floops_data(ods, pulse, nref=0):
+    from scipy.interpolate import interp1d
+    from omfit_classes.omfit_omas_d3d import OMFITd3dcompfile
+
     ods1 = ODS()
     unwrap(magnetics_hardware)(ods1, pulse)
 
@@ -999,11 +1111,48 @@ def magnetics_floops_data(ods, pulse):
             data_norm=1.0,
         )
 
+    # Set reference flux loop to zero before
+    ods[f'magnetics.flux_loop.{nref}.flux.data'] *= 0.0
+
+    for compfile in ['btcomp', 'ccomp', 'icomp']:
+        comp = get_support_file(OMFITd3dcompfile, support_filenames('d3d', compfile, pulse))
+        compshot = -1
+        for shot in comp:
+            if pulse > compshot:
+                compshot = shot
+        for compsig in comp[compshot]:
+            if compsig == 'N1COIL' and pulse > 112962:
+                continue
+            m = mdsvalue('d3d', pulse=pulse, TDI=f'ptdata("{compsig}",{pulse})', treename=None)
+            compsig_data = m.data()
+            compsig_time = m.dim_of(0) / 1000.0
+            for channel in ods['magnetics.flux_loop']:
+                if f'magnetics.flux_loop.{channel}.identifier' in ods1:
+                    sig = ods1[f'magnetics.flux_loop.{channel}.identifier']
+                    sigraw_data = ods[f'magnetics.flux_loop.{channel}.flux.data']
+                    sigraw_time = ods[f'magnetics.flux_loop.{channel}.flux.time']
+                    compsig_data_interp = interp1d(compsig_time, compsig_data, bounds_error=False, fill_value=(0, 0))(sigraw_time)
+                    ods[f'magnetics.flux_loop.{channel}.flux.data'] -= comp[compshot][compsig][sig] * compsig_data_interp
+
+    # fetch uncertainties
+    TDIs = {}
+    for k in ods1['magnetics.flux_loop']:
+        identifier = ods1[f'magnetics.flux_loop.{k}.identifier'].upper()
+        TDIs[identifier] = f'pthead2("{identifier}",{pulse}), __rarray'
+
+    data = mdsvalue('d3d', None, pulse, TDIs).raw()
+    for k in ods1['magnetics.flux_loop']:
+        identifier = ods1[f'magnetics.flux_loop.{k}.identifier'].upper()
+        nt = len(ods[f'magnetics.flux_loop.{k}.flux.data'])
+        ods[f'magnetics.flux_loop.{k}.flux.data_error_upper'] = 10 * abs(data[identifier][3] * data[identifier][4]) * np.ones(nt)
+
 
 @machine_mapping_function(__regression_arguments__, pulse=133221)
 def magnetics_probes_data(ods, pulse):
+    from omfit_classes.omfit_omas_d3d import OMFITd3dcompfile
+
     ods1 = ODS()
-    unwrap(magnetics_hardware)(ods1)
+    unwrap(magnetics_hardware)(ods1, pulse)
 
     with omas_environment(ods, cocosio=1):
         fetch_assign(
@@ -1022,7 +1171,96 @@ def magnetics_probes_data(ods, pulse):
             data_norm=1.0,
         )
 
+    for compfile in ['btcomp', 'ccomp', 'icomp']:
+        comp = get_support_file(OMFITd3dcompfile, support_filenames('d3d', compfile, pulse))
+        compshot = -1
+        for shot in comp:
+            if pulse > compshot:
+                compshot = shot
+        for compsig in comp[compshot]:
+            if compsig == 'N1COIL' and pulse > 112962:
+                continue
+            m = mdsvalue('d3d', pulse=pulse, TDI=f"[ptdata2(\"{compsig}\",{pulse})]", treename=None)
+            compsig_data = m.data()
+
+            compsig_time = m.dim_of(0) / 1000
+            for channel in ods1['magnetics.b_field_pol_probe']:
+                if 'magnetics.b_field_pol_probe.{channel}.identifier' in ods1:
+                    sig = 'magnetics.b_field_pol_probe.{channel}.identifier'
+                    sigraw_data = ods[f'magnetics.b_field_pol_probe.{channel}.field.data']
+                    sigraw_time = ods[f'magnetics.b_field_pol_probe.{channel}.field.time']
+                    compsig_data_interp = np.interp(sigraw_time, compsig_time, compsig_data)
+                    ods[f'magnetics.b_field_pol_probe.{channel}.field.data'] -= comp[compshot][compsig][sig] * compsig_data_interp
+
+    # fetch uncertainties
+    TDIs = {}
+    for k in ods1['magnetics.b_field_pol_probe']:
+        identifier = ods1[f'magnetics.b_field_pol_probe.{k}.identifier'].upper()
+        TDIs[identifier] = f'pthead2("{identifier}",{pulse}), __rarray'
+
+    data = mdsvalue('d3d', None, pulse, TDIs).raw()
+
+    for k in ods1['magnetics.b_field_pol_probe']:
+        identifier = ods1[f'magnetics.b_field_pol_probe.{k}.identifier'].upper()
+        nt = len(ods[f'magnetics.b_field_pol_probe.{k}.field.data'])
+        ods[f'magnetics.b_field_pol_probe.{k}.field.data_error_upper'] = abs(data[identifier][3] * data[identifier][4]) * np.ones(nt) * 10.0
+
+
+@machine_mapping_function(__regression_arguments__, pulse=133221)
+def ip_bt_dflux_data(ods, pulse):
+    r"""
+    Load DIII-D tokamak Ip, Bt, and diamagnetic flux data
+
+    :param ods: ODS instance
+
+    :param pulse: shot number
+    """
+
+    mappings = {'magnetics.ip.0': 'IP', 'tf.b_field_tor_vacuum_r': 'BT', 'magnetics.diamagnetic_flux.0': 'DIAMAG3'}
+
+    with omas_environment(ods, cocosio=1):
+        TDIs = {}
+        for key, val in mappings.items():
+            TDIs[key + '.data'] = f'ptdata("{val}",{pulse})'
+            TDIs[key + '.time'] = f'dim_of(ptdata2("{val}",{pulse}),0)/1000.'
+            TDIs[key + '.data_error_upper'] = f'pthead2("{val}",{pulse}), __rarray'
+
+        data = mdsvalue('d3d', None, pulse, TDIs).raw()
+        for key in TDIs.keys():
+            if 'data_error_upper' in key:
+                nt = len(ods[key[:-12]])
+                ods[key] = abs(data[key][3] * data[key][4]) * np.ones(nt) * 10.0
+            else:
+                ods[key] = data[key]
+
+            if 'magnetics.diamagnetic_flux.0.data' in key:
+                ods[key] *= 1e-3
+
+        ods['tf.b_field_tor_vacuum_r.data'] *= 1.6955
+
+
+# ================================
+@machine_mapping_function(__regression_arguments__, pulse=133221)
+def core_profiles_global_quantities_data(ods, pulse):
+    from scipy.interpolate import interp1d
+
+    ods1 = ODS()
+    unwrap(magnetics_hardware)(ods1)
+
+    with omas_environment(ods, cocosio=1):
+        cp = ods['core_profiles']
+        gq = ods['core_profiles.global_quantities']
+        if 'time' not in cp:
+            m = mdsvalue('d3d', pulse=pulse, TDI="\\TOP.PROFILES.EDENSFIT", treename="ZIPFIT01")
+            cp['time'] = m.dim_of(1) * 1e-3
+        t = cp['time']
+
+        m = mdsvalue('d3d', pulse=pulse, TDI=f"ptdata2(\"VLOOP\",{pulse})", treename=None)
+
+        gq['v_loop'] = interp1d(m.dim_of(0) * 1e-3, m.data(), bounds_error=False, fill_value=np.NaN)(t)
+
 
 # ================================
 if __name__ == '__main__':
+    test_machine_mapping_functions(['ec_launcher_active_hardware'], globals(), locals())
     test_machine_mapping_functions(['thomson_scattering_hardware'], globals(), locals())
